@@ -1,10 +1,82 @@
-#include "image.h"
+#include "imager.h"
 #include "draw.h"
 #include "log.h"
-#include "imagei.h"
-
+#include "imageri.h"
+#include "imrender.h"
 #include <limits.h>
 
+int
+i_ppix_norm(i_img *im, i_img_dim x, i_img_dim y, i_color const *col) {
+  i_color src;
+  i_color work;
+  int dest_alpha;
+  int remains;
+
+  if (!col->channel[3])
+    return 0;
+
+  switch (im->channels) {
+  case 1:
+    work = *col;
+    i_adapt_colors(2, 4, &work, 1);
+    i_gpix(im, x, y, &src);
+    remains = 255 - work.channel[1];
+    src.channel[0] = (src.channel[0] * remains
+		      + work.channel[0] * work.channel[1]) / 255;
+    return i_ppix(im, x, y, &src);
+
+  case 2:
+    work = *col;
+    i_adapt_colors(2, 4, &work, 1);
+    i_gpix(im, x, y, &src);
+    dest_alpha = work.channel[1] + remains * src.channel[1] / 255;
+    if (work.channel[1] == 255) {
+      return i_ppix(im, x, y, &work);
+    }
+    else {
+      src.channel[0] = (work.channel[1] * work.channel[0]
+			+ remains * src.channel[0] * src.channel[1] / 255) / dest_alpha;
+      src.channel[1] = dest_alpha;
+      return i_ppix(im, x, y, &src);
+    }
+
+  case 3:
+    work = *col;
+    i_gpix(im, x, y, &src);
+    remains = 255 - work.channel[3];
+    src.channel[0] = (src.channel[0] * remains
+		      + work.channel[0] * work.channel[3]) / 255;
+    src.channel[1] = (src.channel[1] * remains
+		      + work.channel[1] * work.channel[3]) / 255;
+    src.channel[2] = (src.channel[2] * remains
+		      + work.channel[2] * work.channel[3]) / 255;
+    return i_ppix(im, x, y, &src);
+
+  case 4:
+    work = *col;
+    i_gpix(im, x, y, &src);
+    dest_alpha = work.channel[3] + remains * src.channel[3] / 255;
+    if (work.channel[3] == 255) {
+      return i_ppix(im, x, y, &work);
+    }
+    else {
+      src.channel[0] = (work.channel[3] * work.channel[0]
+			+ remains * src.channel[0] * src.channel[3] / 255) / dest_alpha;
+      src.channel[1] = (work.channel[3] * work.channel[1]
+			+ remains * src.channel[1] * src.channel[3] / 255) / dest_alpha;
+      src.channel[2] = (work.channel[3] * work.channel[2]
+			+ remains * src.channel[2] * src.channel[3] / 255) / dest_alpha;
+      src.channel[3] = dest_alpha;
+      return i_ppix(im, x, y, &src);
+    }
+  }
+  return 0;
+}
+
+static void
+cfill_from_btm(i_img *im, i_fill_t *fill, struct i_bitmap *btm, 
+	       int bxmin, int bxmax, int bymin, int bymax);
+
 void
 i_mmarray_cr(i_mmarray *ar,int l) {
   int i;
@@ -54,64 +126,6 @@ i_mmarray_render(i_img *im,i_mmarray *ar,i_color *val) {
   for(i=0;i<ar->lines;i++) if (ar->data[i].max!=-1) for(x=ar->data[i].min;x<ar->data[i].max;x++) i_ppix(im,x,i,val);
 }
 
-void
-i_mmarray_render_fill(i_img *im,i_mmarray *ar,i_fill_t *fill) {
-  int x, w, y;
-  if (im->bits == i_8_bits && fill->fill_with_color) {
-    i_color *line = mymalloc(sizeof(i_color) * im->xsize); /* checked 5jul05 tonyc */
-    i_color *work = NULL;
-    if (fill->combine)
-      work = mymalloc(sizeof(i_color) * im->xsize); /* checked 5jul05 tonyc */
-    for(y=0;y<ar->lines;y++) {
-      if (ar->data[y].max!=-1) {
-        x = ar->data[y].min;
-        w = ar->data[y].max-ar->data[y].min;
-
-        if (fill->combine) {
-          i_glin(im, x, x+w, y, line);
-          (fill->fill_with_color)(fill, x, y, w, im->channels, work);
-          (fill->combine)(line, work, im->channels, w);
-        }
-        else {
-          (fill->fill_with_color)(fill, x, y, w, im->channels, line);
-        }
-        i_plin(im, x, x+w, y, line);
-      }
-    }
-  
-    myfree(line);
-    if (work)
-      myfree(work);
-  }
-  else {
-    i_fcolor *line = mymalloc(sizeof(i_fcolor) * im->xsize); /* checked 5jul05 tonyc */
-    i_fcolor *work = NULL;
-    if (fill->combinef)
-      work = mymalloc(sizeof(i_fcolor) * im->xsize); /* checked 5jul05 tonyc */
-    for(y=0;y<ar->lines;y++) {
-      if (ar->data[y].max!=-1) {
-        x = ar->data[y].min;
-        w = ar->data[y].max-ar->data[y].min;
-
-        if (fill->combinef) {
-          i_glinf(im, x, x+w, y, line);
-          (fill->fill_with_fcolor)(fill, x, y, w, im->channels, work);
-          (fill->combinef)(line, work, im->channels, w);
-        }
-        else {
-          (fill->fill_with_fcolor)(fill, x, y, w, im->channels, line);
-        }
-        i_plinf(im, x, x+w, y, line);
-      }
-    }
-  
-    myfree(line);
-    if (work)
-      myfree(work);
-  }
-}
-
-
 static
 void
 i_arcdraw(int x1, int y1, int x2, int y2, i_mmarray *ar) {
@@ -201,8 +215,20 @@ i_arc_hlines(i_int_hlines *hlines,int x,int y,float rad,float d1,float d2) {
   }
 }
 
+/*
+=item i_arc(im, x, y, rad, d1, d2, color)
+
+=category Drawing
+=synopsis i_arc(im, 50, 50, 20, 45, 135, &color);
+
+Fills an arc centered at (x,y) with radius I<rad> covering the range
+of angles in degrees from d1 to d2, with the color.
+
+=cut
+*/
+
 void
-i_arc(i_img *im,int x,int y,float rad,float d1,float d2,i_color *val) {
+i_arc(i_img *im,int x,int y,float rad,float d1,float d2,const i_color *val) {
   i_int_hlines hlines;
 
   i_int_init_hlines_img(&hlines, im);
@@ -214,6 +240,18 @@ i_arc(i_img *im,int x,int y,float rad,float d1,float d2,i_color *val) {
   i_int_hlines_destroy(&hlines);
 }
 
+/*
+=item i_arc_cfill(im, x, y, rad, d1, d2, fill)
+
+=category Drawing
+=synopsis i_arc_cfill(im, 50, 50, 35, 90, 135, fill);
+
+Fills an arc centered at (x,y) with radius I<rad> covering the range
+of angles in degrees from d1 to d2, with the fill object.
+
+=cut
+*/
+
 #define MIN_CIRCLE_STEPS 8
 #define MAX_CIRCLE_STEPS 360
 
@@ -277,8 +315,9 @@ arc_poly(int *count, double **xvals, double **yvals,
   angle_inc = 2 * PI / steps;
 
   point_count = steps + 5; /* rough */
-  *xvals = mymalloc(point_count * sizeof(double));
-  *yvals = mymalloc(point_count * sizeof(double));
+  /* point_count is always relatively small, so allocation won't overflow */
+  *xvals = mymalloc(point_count * sizeof(double)); /* checked 17feb2005 tonyc */
+  *yvals = mymalloc(point_count * sizeof(double)); /* checked 17feb2005 tonyc */
 
   /* from centre to edge at d1 */
   (*xvals)[0] = x;
@@ -301,9 +340,21 @@ arc_poly(int *count, double **xvals, double **yvals,
   ++*count;
 }
 
+/*
+=item i_arc_aa(im, x, y, rad, d1, d2, color)
+
+=category Drawing
+=synopsis i_arc_aa(im, 50, 50, 35, 90, 135, &color);
+
+Antialias fills an arc centered at (x,y) with radius I<rad> covering
+the range of angles in degrees from d1 to d2, with the color.
+
+=cut
+*/
+
 void
 i_arc_aa(i_img *im, double x, double y, double rad, double d1, double d2,
-	 i_color *val) {
+	 const i_color *val) {
   double *xvals, *yvals;
   int count;
 
@@ -315,6 +366,18 @@ i_arc_aa(i_img *im, double x, double y, double rad, double d1, double d2,
   myfree(yvals);
 }
 
+/*
+=item i_arc_aa_cfill(im, x, y, rad, d1, d2, fill)
+
+=category Drawing
+=synopsis i_arc_aa_cfill(im, 50, 50, 35, 90, 135, fill);
+
+Antialias fills an arc centered at (x,y) with radius I<rad> covering
+the range of angles in degrees from d1 to d2, with the fill object.
+
+=cut
+*/
+
 void
 i_arc_aa_cfill(i_img *im, double x, double y, double rad, double d1, double d2,
 	       i_fill_t *fill) {
@@ -334,9 +397,6 @@ i_arc_aa_cfill(i_img *im, double x, double y, double rad, double d1, double d2,
 
 typedef int frac;
 static  frac float_to_frac(float x) { return (frac)(0.5+x*16.0); }
-static   int frac_sub     (frac x)  { return (x%16); }
-static   int frac_int     (frac x)  { return (x/16); }
-static float frac_to_float(float x) { return (float)x/16.0; }
 
 static 
 void
@@ -345,19 +405,6 @@ polar_to_plane(float cx, float cy, float angle, float radius, frac *x, frac *y)
   *y = float_to_frac(cy+radius*sin(angle));
 }
 
-static
-void
-order_pair(frac *x, frac *y) {
-  frac t = *x;
-  if (t>*y) {
-    *x = *y;
-    *y = t;
-  }
-}
-
-
-
-
 static
 void
 make_minmax_list(i_mmarray *dot, float x, float y, float radius) {
@@ -425,8 +472,19 @@ i_pixel_coverage(i_mmarray *dot, int x, int y) {
   return cnt;
 }
 
+/*
+=item i_circle_aa(im, x, y, rad, color)
+
+=category Drawing
+=synopsis i_circle_aa(im, 50, 50, 45, &color);
+
+Antialias fills a circle centered at (x,y) for radius I<rad> with
+color.
+
+=cut
+*/
 void
-i_circle_aa(i_img *im, float x, float y, float rad, i_color *val) {
+i_circle_aa(i_img *im, float x, float y, float rad, const i_color *val) {
   i_mmarray dot;
   i_color temp;
   int ly;
@@ -468,13 +526,531 @@ i_circle_aa(i_img *im, float x, float y, float rad, i_color *val) {
   i_mmarray_dst(&dot);
 }
 
+/*
+=item i_circle_out(im, x, y, r, col)
 
+=category Drawing
+=synopsis i_circle_out(im, 50, 50, 45, &color);
 
+Draw a circle outline centered at (x,y) with radius r,
+non-anti-aliased.
 
+Parameters:
 
+=over
+
+=item *
+
+(x, y) - the center of the circle
+
+=item *
+
+r - the radius of the circle in pixels, must be non-negative
+
+=back
+
+Returns non-zero on success.
+
+Implementation:
+
+=cut
+*/
+
+int
+i_circle_out(i_img *im, i_img_dim xc, i_img_dim yc, i_img_dim r,
+	     const i_color *col) {
+  i_img_dim x, y;
+  i_img_dim dx, dy;
+  int error;
+
+  i_clear_error();
+
+  if (r < 0) {
+    i_push_error(0, "circle: radius must be non-negative");
+    return 0;
+  }
+
+  i_ppix(im, xc+r, yc, col);
+  i_ppix(im, xc-r, yc, col);
+  i_ppix(im, xc, yc+r, col);
+  i_ppix(im, xc, yc-r, col);
+
+  x = 0;
+  y = r;
+  dx = 1;
+  dy = -2 * r;
+  error = 1 - r;
+  while (x < y) {
+    if (error >= 0) {
+      --y;
+      dy += 2;
+      error += dy;
+    }
+    ++x;
+    dx += 2;
+    error += dx;
+
+    i_ppix(im, xc + x, yc + y, col);
+    i_ppix(im, xc + x, yc - y, col);
+    i_ppix(im, xc - x, yc + y, col);
+    i_ppix(im, xc - x, yc - y, col);
+    if (x != y) {
+      i_ppix(im, xc + y, yc + x, col);
+      i_ppix(im, xc + y, yc - x, col);
+      i_ppix(im, xc - y, yc + x, col);
+      i_ppix(im, xc - y, yc - x, col);
+    }
+  }
+
+  return 1;
+}
+
+/*
+=item arc_seg(angle)
+
+Convert an angle in degrees into an angle measure we can generate
+simply from the numbers we have when drawing the circle.
+
+=back
+*/
+
+static i_img_dim
+arc_seg(double angle, int scale) {
+  i_img_dim seg = (angle + 45) / 90;
+  double remains = angle - seg * 90; /* should be in the range [-45,45] */
+  int sign = remains < 0 ? -1 : remains ? 1 : 0;
+
+  while (seg > 4)
+    seg -= 4;
+  if (seg == 4 && remains > 0)
+    seg = 0;
+
+  return scale * (seg * 2 + sin(remains * PI/180));
+}
+
+/*
+=item i_arc_out(im, x, y, r, d1, d2, col)
+
+=category Drawing
+=synopsis i_arc_out(im, 50, 50, 45, 45, 135, &color);
+
+Draw an arc outline centered at (x,y) with radius r, non-anti-aliased
+over the angle range d1 through d2 degrees.
+
+Parameters:
+
+=over
+
+=item *
+
+(x, y) - the center of the circle
+
+=item *
+
+r - the radius of the circle in pixels, must be non-negative
+
+=item *
+
+d1, d2 - the range of angles to draw the arc over, in degrees.
+
+=back
+
+Returns non-zero on success.
+
+Implementation:
+
+=cut
+*/
+
+int
+i_arc_out(i_img *im, i_img_dim xc, i_img_dim yc, i_img_dim r,
+	  float d1, float d2, const i_color *col) {
+  i_img_dim x, y;
+  i_img_dim dx, dy;
+  int error;
+  i_img_dim segs[2][2];
+  int seg_count;
+  i_img_dim sin_th;
+  i_img_dim seg_d1, seg_d2;
+  int seg_num;
+  double inv_r;
+  i_img_dim scale = r + 1;
+  i_img_dim seg1 = scale * 2;
+  i_img_dim seg2 = scale * 4;
+  i_img_dim seg3 = scale * 6;
+  i_img_dim seg4 = scale * 8;
+
+  i_clear_error();
+
+  if (r <= 0) {
+    i_push_error(0, "arc: radius must be non-negative");
+    return 0;
+  }
+  if (d1 + 360 <= d2)
+    return i_circle_out(im, xc, yc, r, col);
+
+  if (d1 < 0)
+    d1 += 360 * floor((-d1 + 359) / 360);
+  if (d2 < 0)
+    d2 += 360 * floor((-d2 + 359) / 360);
+  d1 = fmod(d1, 360);
+  d2 = fmod(d2, 360);
+  seg_d1 = arc_seg(d1, scale);
+  seg_d2 = arc_seg(d2, scale);
+  if (seg_d2 < seg_d1) {
+    /* split into two segments */
+    segs[0][0] = 0;
+    segs[0][1] = seg_d2;
+    segs[1][0] = seg_d1;
+    segs[1][1] = seg4;
+    seg_count = 2;
+  }
+  else {
+    segs[0][0] = seg_d1;
+    segs[0][1] = seg_d2;
+    seg_count = 1;
+  }
+
+  for (seg_num = 0; seg_num < seg_count; ++seg_num) {
+    i_img_dim seg_start = segs[seg_num][0];
+    i_img_dim seg_end = segs[seg_num][1];
+    if (seg_start == 0)
+      i_ppix(im, xc+r, yc, col);
+    if (seg_start <= seg1 && seg_end >= seg1)
+      i_ppix(im, xc, yc+r, col);
+    if (seg_start <= seg2 && seg_end >= seg2)
+      i_ppix(im, xc-r, yc, col);
+    if (seg_start <= seg3 && seg_end >= seg3)
+      i_ppix(im, xc, yc-r, col);
+
+    y = 0;
+    x = r;
+    dy = 1;
+    dx = -2 * r;
+    error = 1 - r;
+    while (y < x) {
+      if (error >= 0) {
+	--x;
+	dx += 2;
+	error += dx;
+      }
+      ++y;
+      dy += 2;
+      error += dy;
+      
+      sin_th = y;
+      if (seg_start <= sin_th && seg_end >= sin_th)
+	i_ppix(im, xc + x, yc + y, col);
+      if (seg_start <= seg1 - sin_th && seg_end >= seg1 - sin_th)
+	i_ppix(im, xc + y, yc + x, col);
+
+      if (seg_start <= seg1 + sin_th && seg_end >= seg1 + sin_th)
+	i_ppix(im, xc - y, yc + x, col);
+      if (seg_start <= seg2 - sin_th && seg_end >= seg2 - sin_th)
+	i_ppix(im, xc - x, yc + y, col);
+      
+      if (seg_start <= seg2 + sin_th && seg_end >= seg2 + sin_th)
+	i_ppix(im, xc - x, yc - y, col);
+      if (seg_start <= seg3 - sin_th && seg_end >= seg3 - sin_th)
+	i_ppix(im, xc - y, yc - x, col);
+
+      if (seg_start <= seg3 + sin_th && seg_end >= seg3 + sin_th)
+	i_ppix(im, xc + y, yc - x, col);
+      if (seg_start <= seg4 - sin_th && seg_end >= seg4 - sin_th)
+	i_ppix(im, xc + x, yc - y, col);
+    }
+  }
+
+  return 1;
+}
+
+static double
+cover(i_img_dim r, i_img_dim j) {
+  float rjsqrt = sqrt(r*r - j*j);
+
+  return ceil(rjsqrt) - rjsqrt;
+}
+
+/*
+=item i_circle_out_aa(im, xc, yc, r, col)
+
+=synopsis i_circle_out_aa(im, 50, 50, 45, &color);
+
+Draw a circle outline centered at (x,y) with radius r, anti-aliased.
+
+Parameters:
+
+=over
+
+=item *
+
+(xc, yc) - the center of the circle
+
+=item *
+
+r - the radius of the circle in pixels, must be non-negative
+
+=item *
+
+col - an i_color for the color to draw in.
+
+=back
+
+Returns non-zero on success.
+
+=cut
+
+Based on "Fast Anti-Aliased Circle Generation", Xiaolin Wu, Graphics
+Gems.
+
+I use floating point for I<D> since for large circles the precision of
+a [0,255] value isn't sufficient when approaching the end of the
+octant.
+
+*/
+
+int
+i_circle_out_aa(i_img *im, i_img_dim xc, i_img_dim yc, i_img_dim r, const i_color *col) {
+  i_img_dim i, j;
+  double t;
+  i_color workc = *col;
+  int orig_alpha = col->channel[3];
+
+  i_clear_error();
+  if (r <= 0) {
+    i_push_error(0, "arc: radius must be non-negative");
+    return 0;
+  }
+  i = r;
+  j = 0;
+  t = 0;
+  i_ppix_norm(im, xc+i, yc+j, col);
+  i_ppix_norm(im, xc-i, yc+j, col);
+  i_ppix_norm(im, xc+j, yc+i, col);
+  i_ppix_norm(im, xc+j, yc-i, col);
+
+  while (i > j+1) {
+    double d;
+    int cv, inv_cv;
+    i_color p;
+    int ch;
+    j++;
+    d = cover(r, j);
+    cv = (int)(d * 255 + 0.5);
+    inv_cv = 255-cv;
+    if (d < t) {
+      --i;
+    }
+    if (inv_cv) {
+      workc.channel[3] = orig_alpha * inv_cv / 255;
+      i_ppix_norm(im, xc+i, yc+j, &workc);
+      i_ppix_norm(im, xc-i, yc+j, &workc);
+      i_ppix_norm(im, xc+i, yc-j, &workc);
+      i_ppix_norm(im, xc-i, yc-j, &workc);
+
+      if (i != j) {
+	i_ppix_norm(im, xc+j, yc+i, &workc);
+	i_ppix_norm(im, xc-j, yc+i, &workc);
+	i_ppix_norm(im, xc+j, yc-i, &workc);
+	i_ppix_norm(im, xc-j, yc-i, &workc);
+      }
+    }
+    if (cv && i > j) {
+      workc.channel[3] = orig_alpha * cv / 255;
+      i_ppix_norm(im, xc+i-1, yc+j, &workc);
+      i_ppix_norm(im, xc-i+1, yc+j, &workc);
+      i_ppix_norm(im, xc+i-1, yc-j, &workc);
+      i_ppix_norm(im, xc-i+1, yc-j, &workc);
+
+      if (j != i-1) {
+	i_ppix_norm(im, xc+j, yc+i-1, &workc);
+	i_ppix_norm(im, xc-j, yc+i-1, &workc);
+	i_ppix_norm(im, xc+j, yc-i+1, &workc);
+	i_ppix_norm(im, xc-j, yc-i+1, &workc);
+      }
+    }
+    t = d;
+  }
+
+  return 1;
+}
+
+/*
+=item i_arc_out_aa(im, xc, yc, r, d1, d2, col)
+
+=synopsis i_arc_out_aa(im, 50, 50, 45, 45, 125, &color);
+
+Draw a circle arc outline centered at (x,y) with radius r, from angle
+d1 degrees through angle d2 degrees, anti-aliased.
+
+Parameters:
+
+=over
+
+=item *
+
+(xc, yc) - the center of the circle
+
+=item *
+
+r - the radius of the circle in pixels, must be non-negative
+
+=item *
+
+d1, d2 - the range of angle in degrees to draw the arc through.  If
+d2-d1 >= 360 a full circle is drawn.
+
+=back
+
+Returns non-zero on success.
+
+=cut
+
+Based on "Fast Anti-Aliased Circle Generation", Xiaolin Wu, Graphics
+Gems.
+
+*/
+
+int
+i_arc_out_aa(i_img *im, i_img_dim xc, i_img_dim yc, i_img_dim r, float d1, float d2, const i_color *col) {
+  i_img_dim i, j;
+  double t;
+  i_color workc = *col;
+  i_img_dim segs[2][2];
+  int seg_count;
+  i_img_dim sin_th;
+  i_img_dim seg_d1, seg_d2;
+  int seg_num;
+  int orig_alpha = col->channel[3];
+  i_img_dim scale = r + 1;
+  i_img_dim seg1 = scale * 2;
+  i_img_dim seg2 = scale * 4;
+  i_img_dim seg3 = scale * 6;
+  i_img_dim seg4 = scale * 8;
+
+  i_clear_error();
+  if (r <= 0) {
+    i_push_error(0, "arc: radius must be non-negative");
+    return 0;
+  }
+  if (d1 + 360 <= d2)
+    return i_circle_out_aa(im, xc, yc, r, col);
+
+  if (d1 < 0)
+    d1 += 360 * floor((-d1 + 359) / 360);
+  if (d2 < 0)
+    d2 += 360 * floor((-d2 + 359) / 360);
+  d1 = fmod(d1, 360);
+  d2 = fmod(d2, 360);
+  seg_d1 = arc_seg(d1, scale);
+  seg_d2 = arc_seg(d2, scale);
+  if (seg_d2 < seg_d1) {
+    /* split into two segments */
+    segs[0][0] = 0;
+    segs[0][1] = seg_d2;
+    segs[1][0] = seg_d1;
+    segs[1][1] = seg4;
+    seg_count = 2;
+  }
+  else {
+    segs[0][0] = seg_d1;
+    segs[0][1] = seg_d2;
+    seg_count = 1;
+  }
+
+  for (seg_num = 0; seg_num < seg_count; ++seg_num) {
+    i_img_dim seg_start = segs[seg_num][0];
+    i_img_dim seg_end = segs[seg_num][1];
+
+    i = r;
+    j = 0;
+    t = 0;
+
+    if (seg_start == 0)
+      i_ppix_norm(im, xc+i, yc+j, col);
+    if (seg_start <= seg1 && seg_end >= seg1)
+      i_ppix_norm(im, xc+j, yc+i, col);
+    if (seg_start <= seg2 && seg_end >= seg2)
+      i_ppix_norm(im, xc-i, yc+j, col);
+    if (seg_start <= seg3 && seg_end >= seg3)
+      i_ppix_norm(im, xc+j, yc-i, col);
+    
+    while (i > j+1) {
+      int cv, inv_cv;
+      i_color p;
+      int ch;
+      double d;
+      j++;
+      d = cover(r, j);
+      cv = (int)(d * 255 + 0.5);
+      inv_cv = 255-cv;
+      if (d < t) {
+	--i;
+      }
+      sin_th = j;
+      if (inv_cv) {
+	workc.channel[3] = orig_alpha * inv_cv / 255;
+
+	if (seg_start <= sin_th && seg_end >= sin_th)
+	  i_ppix_norm(im, xc+i, yc+j, &workc);
+	if (seg_start <= seg2 - sin_th && seg_end >= seg2 - sin_th)
+	  i_ppix_norm(im, xc-i, yc+j, &workc);
+	if (seg_start <= seg4 - sin_th && seg_end >= seg4 - sin_th)
+	  i_ppix_norm(im, xc+i, yc-j, &workc);
+	if (seg_start <= seg2 + sin_th && seg_end >= seg2 + sin_th)
+	  i_ppix_norm(im, xc-i, yc-j, &workc);
+	
+	if (i != j) {
+	  if (seg_start <= seg1 - sin_th && seg_end >= seg1 - sin_th)
+	    i_ppix_norm(im, xc+j, yc+i, &workc);
+	  if (seg_start <= seg1 + sin_th && seg_end >= seg1 + sin_th)
+	    i_ppix_norm(im, xc-j, yc+i, &workc);
+	  if (seg_start <= seg3 + sin_th && seg_end >= seg3 + sin_th)
+	    i_ppix_norm(im, xc+j, yc-i, &workc);
+	  if (seg_start <= seg3 - sin_th && seg_end >= seg3 - sin_th)
+	    i_ppix_norm(im, xc-j, yc-i, &workc);
+	}
+      }
+      if (cv && i > j) {
+	workc.channel[3] = orig_alpha * cv / 255;
+	if (seg_start <= sin_th && seg_end >= sin_th)
+	  i_ppix_norm(im, xc+i-1, yc+j, &workc);
+	if (seg_start <= seg2 - sin_th && seg_end >= seg2 - sin_th)
+	  i_ppix_norm(im, xc-i+1, yc+j, &workc);
+	if (seg_start <= seg4 - sin_th && seg_end >= seg4 - sin_th)
+	  i_ppix_norm(im, xc+i-1, yc-j, &workc);
+	if (seg_start <= seg2 + sin_th && seg_end >= seg2 + sin_th)
+	  i_ppix_norm(im, xc-i+1, yc-j, &workc);
+	
+	if (seg_start <= seg1 - sin_th && seg_end >= seg1 - sin_th)
+	  i_ppix_norm(im, xc+j, yc+i-1, &workc);
+	if (seg_start <= seg1 + sin_th && seg_end >= seg1 + sin_th)
+	  i_ppix_norm(im, xc-j, yc+i-1, &workc);
+	if (seg_start <= seg3 + sin_th && seg_end >= seg3 + sin_th)
+	  i_ppix_norm(im, xc+j, yc-i+1, &workc);
+	if (seg_start <= seg3 - sin_th && seg_end >= seg3 - sin_th)
+	  i_ppix_norm(im, xc-j, yc-i+1, &workc);
+      }
+      t = d;
+    }
+  }
+
+  return 1;
+}
+
+/*
+=item i_box(im, x1, y1, x2, y2, color)
+
+=category Drawing
+=synopsis i_box(im, 0, 0, im->xsize-1, im->ysize-1, &color).
+
+Outlines the box from (x1,y1) to (x2,y2) inclusive with I<color>.
+
+=cut
+*/
 
 void
-i_box(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
+i_box(i_img *im,int x1,int y1,int x2,int y2,const i_color *val) {
   int x,y;
   mm_log((1,"i_box(im* 0x%x,x1 %d,y1 %d,x2 %d,y2 %d,val 0x%x)\n",im,x1,y1,x2,y2,val));
   for(x=x1;x<x2+1;x++) {
@@ -487,15 +1063,38 @@ i_box(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
   }
 }
 
+/*
+=item i_box_filled(im, x1, y1, x2, y2, color)
+
+=category Drawing
+=synopsis i_box_filled(im, 0, 0, im->xsize-1, im->ysize-1, &color);
+
+Fills the box from (x1,y1) to (x2,y2) inclusive with color.
+
+=cut
+*/
+
 void
-i_box_filled(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
+i_box_filled(i_img *im,int x1,int y1,int x2,int y2, const i_color *val) {
   int x,y;
   mm_log((1,"i_box_filled(im* 0x%x,x1 %d,y1 %d,x2 %d,y2 %d,val 0x%x)\n",im,x1,y1,x2,y2,val));
   for(x=x1;x<x2+1;x++) for (y=y1;y<y2+1;y++) i_ppix(im,x,y,val);
 }
 
+/*
+=item i_box_cfill(im, x1, y1, x2, y2, fill)
+
+=category Drawing
+=synopsis i_box_cfill(im, 0, 0, im->xsize-1, im->ysize-1, fill);
+
+Fills the box from (x1,y1) to (x2,y2) inclusive with fill.
+
+=cut
+*/
+
 void
 i_box_cfill(i_img *im,int x1,int y1,int x2,int y2,i_fill_t *fill) {
+  i_render r;
   mm_log((1,"i_box_cfill(im* 0x%x,x1 %d,y1 %d,x2 %d,y2 %d,fill 0x%x)\n",im,x1,y1,x2,y2,fill));
 
   ++x2;
@@ -509,54 +1108,20 @@ i_box_cfill(i_img *im,int x1,int y1,int x2,int y2,i_fill_t *fill) {
     y2 = im->ysize-1;
   if (x1 >= x2 || y1 > y2)
     return;
-  if (im->bits == i_8_bits && fill->fill_with_color) {
-    i_color *line = mymalloc(sizeof(i_color) * (x2 - x1)); /* checked 5jul05 tonyc */
-    i_color *work = NULL;
-    if (fill->combine)
-      work = mymalloc(sizeof(i_color) * (x2-x1)); /* checked 5jul05 tonyc */
-    while (y1 <= y2) {
-      if (fill->combine) {
-        i_glin(im, x1, x2, y1, line);
-        (fill->fill_with_color)(fill, x1, y1, x2-x1, im->channels, work);
-        (fill->combine)(line, work, im->channels, x2-x1);
-      }
-      else {
-        (fill->fill_with_color)(fill, x1, y1, x2-x1, im->channels, line);
-      }
-      i_plin(im, x1, x2, y1, line);
-      ++y1;
-    }
-    myfree(line);
-    if (work)
-      myfree(work);
-  }
-  else {
-    i_fcolor *line = mymalloc(sizeof(i_fcolor) * (x2 - x1)); /* checked 5jul05 tonyc */
-    i_fcolor *work;
-    work = mymalloc(sizeof(i_fcolor) * (x2 - x1)); /* checked 5jul05 tonyc */
 
-    while (y1 <= y2) {
-      if (fill->combine) {
-        i_glinf(im, x1, x2, y1, line);
-        (fill->fill_with_fcolor)(fill, x1, y1, x2-x1, im->channels, work);
-        (fill->combinef)(line, work, im->channels, x2-x1);
-      }
-      else {
-        (fill->fill_with_fcolor)(fill, x1, y1, x2-x1, im->channels, line);
-      }
-      i_plinf(im, x1, x2, y1, line);
-      ++y1;
-    }
-    myfree(line);
-    if (work)
-      myfree(work);
+  i_render_init(&r, im, x2-x1);
+  while (y1 <= y2) {
+    i_render_fill(&r, x1, y1, x2-x1, NULL, fill);
+    ++y1;
   }
+  i_render_done(&r);
 }
 
-
 /* 
 =item i_line(im, x1, y1, x2, y2, val, endp)
 
+=category Drawing
+
 Draw a line to image using bresenhams linedrawing algorithm
 
    im   - image to draw to
@@ -571,7 +1136,7 @@ Draw a line to image using bresenhams linedrawing algorithm
 */
 
 void
-i_line(i_img *im, int x1, int y1, int x2, int y2, i_color *val, int endp) {
+i_line(i_img *im, int x1, int y1, int x2, int y2, const i_color *val, int endp) {
   int x, y;
   int dx, dy;
   int p;
@@ -672,94 +1237,20 @@ i_line_dda(i_img *im, int x1, int y1, int x2, int y2, i_color *val) {
   }
 }
 
+/*
+=item i_line_aa(im, x1, x2, y1, y2, color, endp)
 
+=category Drawing
 
+Antialias draws a line from (x1,y1) to (x2, y2) in color.
 
+The point (x2, y2) is drawn only if endp is set.
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
+=cut
+*/
 
 void
-i_line_aa3(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
-  i_color tval;
-  float alpha;
-  float dsec,dfrac;
-  int temp,dx,dy,isec,ch;
-
-  mm_log((1,"i_line_aa(im* 0x%x,x1 %d,y1 %d,x2 %d,y2 %d,val 0x%x)\n",im,x1,y1,x2,y2,val));
-
-  dy=y2-y1;
-  dx=x2-x1;
-
-  if (abs(dx)>abs(dy)) { /* alpha < 1 */
-    if (x2<x1) { temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp; }
-    alpha=(float)(y2-y1)/(float)(x2-x1);
-
-    dsec=y1;
-    while(x1<=x2) {
-      isec=(int)dsec;
-      dfrac=dsec-isec;
-      /*      dfrac=1-(1-dfrac)*(1-dfrac); */
-      /* This is something we can play with to try to get better looking lines */
-
-      i_gpix(im,x1,isec,&tval);
-      for(ch=0;ch<im->channels;ch++) tval.channel[ch]=(unsigned char)(dfrac*(float)tval.channel[ch]+(1-dfrac)*(float)val->channel[ch]);
-      i_ppix(im,x1,isec,&tval);
-      
-      i_gpix(im,x1,isec+1,&tval);
-      for(ch=0;ch<im->channels;ch++) tval.channel[ch]=(unsigned char)((1-dfrac)*(float)tval.channel[ch]+dfrac*(float)val->channel[ch]);
-      i_ppix(im,x1,isec+1,&tval);
-      
-      dsec+=alpha;
-      x1++;
-    }
-  } else {
-    if (y2<y1) { temp=y1; y1=y2; y2=temp; temp=x1; x1=x2; x2=temp; }
-    alpha=(float)(x2-x1)/(float)(y2-y1);
-    dsec=x1;
-    while(y1<=y2) {
-      isec=(int)dsec;
-      dfrac=dsec-isec;
-      /*      dfrac=sqrt(dfrac); */
-      /* This is something we can play with */
-      i_gpix(im,isec,y1,&tval);
-      for(ch=0;ch<im->channels;ch++) tval.channel[ch]=(unsigned char)(dfrac*(float)tval.channel[ch]+(1-dfrac)*(float)val->channel[ch]);
-      i_ppix(im,isec,y1,&tval);
-
-      i_gpix(im,isec+1,y1,&tval);
-      for(ch=0;ch<im->channels;ch++) tval.channel[ch]=(unsigned char)((1-dfrac)*(float)tval.channel[ch]+dfrac*(float)val->channel[ch]);
-      i_ppix(im,isec+1,y1,&tval);
-
-      dsec+=alpha;
-      y1++;
-    }
-  }
-}
-
-
-
-
-void
-i_line_aa(i_img *im, int x1, int y1, int x2, int y2, i_color *val, int endp) {
+i_line_aa(i_img *im, int x1, int y1, int x2, int y2, const i_color *val, int endp) {
   int x, y;
   int dx, dy;
   int p;
@@ -903,7 +1394,7 @@ perm(int n,int k) {
    to get a new level - this may lead to errors who knows lets test it */
 
 void
-i_bezier_multi(i_img *im,int l,double *x,double *y,i_color *val) {
+i_bezier_multi(i_img *im,int l,const double *x,const double *y, const i_color *val) {
   double *bzcoef;
   double t,cx,cy;
   int k,i;
@@ -1002,33 +1493,46 @@ crdata(int left,int right,int dadl,int dadr,int y, int dir) {
 
 /* i_ccomp compares two colors and gives true if they are the same */
 
+typedef int (*ff_cmpfunc)(i_color const *c1, i_color const *c2, int channels);
+
 static int
-i_ccomp(i_color *val1,i_color *val2,int ch) {
+i_ccomp_normal(i_color const *val1, i_color const *val2, int ch) {
   int i;
-  for(i=0;i<ch;i++) if (val1->channel[i] !=val2->channel[i]) return 0;
+  for(i = 0; i < ch; i++) 
+    if (val1->channel[i] !=val2->channel[i])
+      return 0;
   return 1;
 }
 
+static int
+i_ccomp_border(i_color const *val1, i_color const *val2, int ch) {
+  int i;
+  for(i = 0; i < ch; i++) 
+    if (val1->channel[i] !=val2->channel[i])
+      return 1;
+  return 0;
+}
 
 static int
-i_lspan(i_img *im, int seedx, int seedy, i_color *val) {
+i_lspan(i_img *im, int seedx, int seedy, i_color const *val, ff_cmpfunc cmpfunc) {
   i_color cval;
   while(1) {
     if (seedx-1 < 0) break;
     i_gpix(im,seedx-1,seedy,&cval);
-    if (!i_ccomp(val,&cval,im->channels)) break;
+    if (!cmpfunc(val,&cval,im->channels)) 
+      break;
     seedx--;
   }
   return seedx;
 }
 
 static int
-i_rspan(i_img *im, int seedx, int seedy, i_color *val) {
+i_rspan(i_img *im, int seedx, int seedy, i_color const *val, ff_cmpfunc cmpfunc) {
   i_color cval;
   while(1) {
     if (seedx+1 > im->xsize-1) break;
     i_gpix(im,seedx+1,seedy,&cval);
-    if (!i_ccomp(val,&cval,im->channels)) break;
+    if (!cmpfunc(val,&cval,im->channels)) break;
     seedx++;
   }
   return seedx;
@@ -1068,7 +1572,7 @@ i_rspan(i_img *im, int seedx, int seedy, i_color *val) {
 #define SET(x,y) btm_set(btm,x,y)
 
 /* INSIDE returns true if pixel is correct color and we haven't set it before. */
-#define INSIDE(x,y) ((!btm_test(btm,x,y) && ( i_gpix(im,x,y,&cval),i_ccomp(&val,&cval,channels)  ) ))
+#define INSIDE(x,y, seed) ((!btm_test(btm,x,y) && ( i_gpix(im,x,y,&cval),cmpfunc(seed,&cval,channels)  ) ))
 
 
 
@@ -1076,15 +1580,8 @@ i_rspan(i_img *im, int seedx, int seedy, i_color *val) {
 
 static struct i_bitmap *
 i_flood_fill_low(i_img *im,int seedx,int seedy,
-                 int *bxminp, int *bxmaxp, int *byminp, int *bymaxp) {
-
-  /*
-    int lx,rx;
-    int y;
-    int direction;
-    int dadLx,dadRx;
-    int wasIn=0;
-  */
+                 int *bxminp, int *bxmaxp, int *byminp, int *bymaxp,
+		 i_color const *seed, ff_cmpfunc cmpfunc) {
   int ltx, rtx;
   int tx = 0;
 
@@ -1097,7 +1594,7 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
   struct i_bitmap *btm;
 
   int channels,xsize,ysize;
-  i_color cval,val;
+  i_color cval;
 
   channels = im->channels;
   xsize    = im->xsize;
@@ -1106,12 +1603,9 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
   btm = btm_new(xsize, ysize);
   st  = llist_new(100, sizeof(struct stack_element*));
 
-  /* Get the reference color */
-  i_gpix(im, seedx, seedy, &val);
-
   /* Find the starting span and fill it */
-  ltx = i_lspan(im, seedx, seedy, &val);
-  rtx = i_rspan(im, seedx, seedy, &val);
+  ltx = i_lspan(im, seedx, seedy, seed, cmpfunc);
+  rtx = i_rspan(im, seedx, seedy, seed, cmpfunc);
   for(tx=ltx; tx<=rtx; tx++) SET(tx, seedy);
 
   ST_PUSH(ltx, rtx, ltx, rtx, seedy+1,  1);
@@ -1136,10 +1630,10 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
 
 
     x = lx+1;
-    if ( lx >= 0 && (wasIn = INSIDE(lx, y)) ) {
+    if ( lx >= 0 && (wasIn = INSIDE(lx, y, seed)) ) {
       SET(lx, y);
       lx--;
-      while(INSIDE(lx, y) && lx > 0) {
+      while(INSIDE(lx, y, seed) && lx > 0) {
 	SET(lx,y);
 	lx--;
       }
@@ -1150,7 +1644,7 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
       /*  printf("x=%d\n",x); */
       if (wasIn) {
 	
-	if (INSIDE(x, y)) {
+	if (INSIDE(x, y, seed)) {
 	  /* case 1: was inside, am still inside */
 	  SET(x,y);
 	} else {
@@ -1163,7 +1657,7 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
 	}
       } else {
 	if (x > rx) goto EXT;
-	if (INSIDE(x, y)) {
+	if (INSIDE(x, y, seed)) {
 	  SET(x, y);
 	  /* case 3: Wasn't inside, am now: just found the start of a new run */
 	  wasIn = 1;
@@ -1192,14 +1686,26 @@ i_flood_fill_low(i_img *im,int seedx,int seedy,
   return btm;
 }
 
+/*
+=item i_flood_fill(im, seedx, seedy, color)
 
+=category Drawing
+=synopsis i_flood_fill(im, 50, 50, &color);
 
+Flood fills the 4-connected region starting from the point (seedx,
+seedy) with I<color>.
+
+Returns false if (seedx, seedy) are outside the image.
+
+=cut
+*/
 
 undef_int
-i_flood_fill(i_img *im, int seedx, int seedy, i_color *dcol) {
+i_flood_fill(i_img *im, int seedx, int seedy, const i_color *dcol) {
   int bxmin, bxmax, bymin, bymax;
   struct i_bitmap *btm;
   int x, y;
+  i_color val;
 
   i_clear_error();
   if (seedx < 0 || seedx >= im->xsize ||
@@ -1208,7 +1714,11 @@ i_flood_fill(i_img *im, int seedx, int seedy, i_color *dcol) {
     return 0;
   }
 
-  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax);
+  /* Get the reference color */
+  i_gpix(im, seedx, seedy, &val);
+
+  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax,
+			 &val, i_ccomp_normal);
 
   for(y=bymin;y<=bymax;y++)
     for(x=bxmin;x<=bxmax;x++)
@@ -1218,14 +1728,25 @@ i_flood_fill(i_img *im, int seedx, int seedy, i_color *dcol) {
   return 1;
 }
 
+/*
+=item i_flood_cfill(im, seedx, seedy, fill)
 
+=category Drawing
+=synopsis i_flood_cfill(im, 50, 50, fill);
+
+Flood fills the 4-connected region starting from the point (seedx,
+seedy) with I<fill>.
+
+Returns false if (seedx, seedy) are outside the image.
+
+=cut
+*/
 
 undef_int
 i_flood_cfill(i_img *im, int seedx, int seedy, i_fill_t *fill) {
   int bxmin, bxmax, bymin, bymax;
   struct i_bitmap *btm;
-  int x, y;
-  int start;
+  i_color val;
 
   i_clear_error();
   
@@ -1235,81 +1756,121 @@ i_flood_cfill(i_img *im, int seedx, int seedy, i_fill_t *fill) {
     return 0;
   }
 
-  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax);
+  /* Get the reference color */
+  i_gpix(im, seedx, seedy, &val);
 
-  if (im->bits == i_8_bits && fill->fill_with_color) {
-    /* bxmax/bxmin are inside the image, hence this won't overflow */
-    i_color *line = mymalloc(sizeof(i_color) * (bxmax - bxmin)); /* checked 5jul05 tonyc */
-    i_color *work = NULL;
-    if (fill->combine)
-      work = mymalloc(sizeof(i_color) * (bxmax - bxmin)); /* checked 5jul05 tonyc */
+  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax,
+			 &val, i_ccomp_normal);
 
-    for(y=bymin; y<=bymax; y++) {
-      x = bxmin;
-      while (x < bxmax) {
-        while (x < bxmax && !btm_test(btm, x, y)) {
-          ++x;
-        }
-        if (btm_test(btm, x, y)) {
-          start = x;
-          while (x < bxmax && btm_test(btm, x, y)) {
-            ++x;
-          }
-          if (fill->combine) {
-            i_glin(im, start, x, y, line);
-            (fill->fill_with_color)(fill, start, y, x-start, im->channels, 
-                                    work);
-            (fill->combine)(line, work, im->channels, x-start);
-          }
-          else {
-            (fill->fill_with_color)(fill, start, y, x-start, im->channels, 
-                                    line);
-          }
-          i_plin(im, start, x, y, line);
-        }
-      }
-    }
-    myfree(line);
-    if (work)
-      myfree(work);
-  }
-  else {
-    /* bxmax/bxmin are inside the image, hence this won't overflow */
-    i_fcolor *line = mymalloc(sizeof(i_fcolor) * (bxmax - bxmin)); /* checked 5jul05 tonyc */
-    i_fcolor *work = NULL;
-    if (fill->combinef)
-      work = mymalloc(sizeof(i_fcolor) * (bxmax - bxmin)); /* checked 5jul05 tonyc */
-    
-    for(y=bymin;y<=bymax;y++) {
-      x = bxmin;
-      while (x < bxmax) {
-        while (x < bxmax && !btm_test(btm, x, y)) {
-          ++x;
-        }
-        if (btm_test(btm, x, y)) {
-          start = x;
-          while (x < bxmax && btm_test(btm, x, y)) {
-            ++x;
-          }
-          if (fill->combinef) {
-            i_glinf(im, start, x, y, line);
-            (fill->fill_with_fcolor)(fill, start, y, x-start, im->channels, 
-                                    work);
-            (fill->combinef)(line, work, im->channels, x-start);
-          }
-          else {
-            (fill->fill_with_fcolor)(fill, start, y, x-start, im->channels, 
-                                    line);
-          }
-          i_plinf(im, start, x, y, line);
-        }
-      }
-    }
-    myfree(line);
-    if (work)
-      myfree(work);
-  }
+  cfill_from_btm(im, fill, btm, bxmin, bxmax, bymin, bymax);
 
   btm_destroy(btm);
   return 1;
 }
+
+/*
+=item i_flood_fill_border(im, seedx, seedy, color, border)
+
+=category Drawing
+=synopsis i_flood_fill_border(im, 50, 50, &color, &border);
+
+Flood fills the 4-connected region starting from the point (seedx,
+seedy) with I<color>, fill stops when the fill reaches a pixels with
+color I<border>.
+
+Returns false if (seedx, seedy) are outside the image.
+
+=cut
+*/
+
+undef_int
+i_flood_fill_border(i_img *im, int seedx, int seedy, const i_color *dcol,
+		    const i_color *border) {
+  int bxmin, bxmax, bymin, bymax;
+  struct i_bitmap *btm;
+  int x, y;
+
+  i_clear_error();
+  if (seedx < 0 || seedx >= im->xsize ||
+      seedy < 0 || seedy >= im->ysize) {
+    i_push_error(0, "i_flood_cfill: Seed pixel outside of image");
+    return 0;
+  }
+
+  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax,
+			 border, i_ccomp_border);
+
+  for(y=bymin;y<=bymax;y++)
+    for(x=bxmin;x<=bxmax;x++)
+      if (btm_test(btm,x,y)) 
+	i_ppix(im,x,y,dcol);
+  btm_destroy(btm);
+  return 1;
+}
+
+/*
+=item i_flood_cfill_border(im, seedx, seedy, fill, border)
+
+=category Drawing
+=synopsis i_flood_cfill_border(im, 50, 50, fill, border);
+
+Flood fills the 4-connected region starting from the point (seedx,
+seedy) with I<fill>, the fill stops when it reaches pixels of color
+I<border>.
+
+Returns false if (seedx, seedy) are outside the image.
+
+=cut
+*/
+
+undef_int
+i_flood_cfill_border(i_img *im, int seedx, int seedy, i_fill_t *fill,
+		     const i_color *border) {
+  int bxmin, bxmax, bymin, bymax;
+  struct i_bitmap *btm;
+
+  i_clear_error();
+  
+  if (seedx < 0 || seedx >= im->xsize ||
+      seedy < 0 || seedy >= im->ysize) {
+    i_push_error(0, "i_flood_cfill_border: Seed pixel outside of image");
+    return 0;
+  }
+
+  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax,
+			 border, i_ccomp_border);
+
+  cfill_from_btm(im, fill, btm, bxmin, bxmax, bymin, bymax);
+
+  btm_destroy(btm);
+
+  return 1;
+}
+
+static void
+cfill_from_btm(i_img *im, i_fill_t *fill, struct i_bitmap *btm, 
+	       int bxmin, int bxmax, int bymin, int bymax) {
+  int x, y;
+  int start;
+
+  i_render r;
+
+  i_render_init(&r, im, bxmax - bxmin + 1);
+
+  for(y=bymin; y<=bymax; y++) {
+    x = bxmin;
+    while (x <= bxmax) {
+      while (x <= bxmax && !btm_test(btm, x, y)) {
+	++x;
+      }
+      if (btm_test(btm, x, y)) {
+	start = x;
+	while (x <= bxmax && btm_test(btm, x, y)) {
+	  ++x;
+	}
+	i_render_fill(&r, start, y, x-start, NULL, fill);
+      }
+    }
+  }
+  i_render_done(&r);
+}
