@@ -1,5 +1,8 @@
 package Imager::Matrix2d;
 use strict;
+use vars qw($VERSION);
+
+$VERSION = "1.008";
 
 =head1 NAME
 
@@ -91,7 +94,9 @@ sub rotate {
     return undef;
   }
 
-  if ($opts{'x'} && $opts{'y'}) {
+  if ($opts{'x'} || $opts{'y'}) {
+    $opts{'x'} ||= 0;
+    $opts{'y'} ||= 0;
     return $class->translate('x'=>-$opts{'x'}, 'y'=>-$opts{'y'})
       * $class->rotate(radians=>$angle)
         * $class->translate('x'=>$opts{'x'}, 'y'=>$opts{'y'});
@@ -107,24 +112,35 @@ sub rotate {
 
 =item translate(x=>$dx, y=>$dy)
 
+=item translate(x=>$dx)
+
+=item translate(y=>$dy)
+
 Translates by the specify amounts.
 
 =cut
+
 sub translate {
   my ($class, %opts) = @_;
 
-  if (defined $opts{'x'} && defined $opts{'y'}) {
-    return bless [ 1, 0, $opts{'x'},
-                   0, 1, $opts{'y'},
+  if (defined $opts{'x'} || defined $opts{'y'}) {
+    my $x = $opts{'x'} || 0;
+    my $y = $opts{'y'} || 0;
+    return bless [ 1, 0, $x,
+                   0, 1, $y,
                    0, 0, 1 ], $class;
   }
 
-  $Imager::ERRSTR = 'x and y parameters required';
+  $Imager::ERRSTR = 'x or y parameter required';
   return undef;
 }
 
 =item shear(x=>$sx, y=>$sy)
 
+=item shear(x=>$sx)
+
+=item shear(y=>$sy)
+
 Shear by the given amounts.
 
 =cut
@@ -286,6 +302,7 @@ newline.
 I tried to make it fairly nicely formatted.  You might disagree :)
 
 =cut
+
 sub _string {
   my ($m) = @_;
 
@@ -343,6 +360,8 @@ You can import these methods with:
 
 =item m2d_reflect()
 
+=item m2d_scale()
+
 =back
 
 =cut
@@ -385,6 +404,6 @@ Needs a way to invert matrixes.
 
 Imager(3), Imager::Font(3)
 
-http://www.eecs.umich.edu/~addi/perl/Imager/
+http://imager.perl.org/
 
 =cut
